
type lex_parser =
  {
  (* demands *)
  demand_line   : unit   -> unit;
  demand_word   : string -> unit;
  demand_mono   : char   -> unit;
  demand_poly   : string -> unit;
  demand_name   : unit   -> string;
  demand_not    : (string -> bool) -> string;
  demand_string : unit   -> string;
  demand_char   : unit   -> char;
  demand_int    : unit   -> int;
  (* promises *)
  promised_word : string -> bool;
  promised_mono : char   -> bool;
  promised_name : unit   -> bool;
  promised_int  : unit   -> bool;
  (* grants *)
  granted_word  : string -> bool;
  granted_mono  : char   -> bool;
  granted_poly  : string -> bool;
  }

type position =
  {line: int; text: string; start: int; stop: int}

type demand =
  | Word of string
  | Mono of char
  | Name
  | String
  | Char
  | Int
  
exception Demand_denied of position * demand

type kind = K_Word | K_Mono | K_Number


(* reserved words*)

exception Word_reserved of position

let sorted_array_mem a x =
  let rec helper l r =
    l < r &&
    let p = (r + l) / 2 in
    if a.(p) < x then helper (succ p) r
    else if a.(p) > x then helper l p
    else true
  in helper 0 (Array.length a)


(* constructor *)

let make file =

  let str = ref ""
  and len = ref 0
  and row = ref 0
  and   l = ref 0
  and   r = ref 0
  and num = ref 0
  and lex = ref K_Mono
  in

  let read_text () =
    str := input_line file;
    len := String.length !str;
    incr row; l := 0; r := 0
  in

  let read_next () =
    let blank_space () =
      if !r < !len then let c = !str.[!r] in (c=' ' or c='\t')
      else (read_text (); decr r; true)
    and upper_lower () =
      !r < !len && let c = !str.[!r] in
      (c>='a' && c<='z' or c>='A' && c<='Z') 
    and alpha_num () =
      !r < !len && let c = !str.[!r] in
      (c>='a' && c<='z' or c>='A' && c<='Z' or c='_' or c>='0' && c<='9')
    and ascii_num () =
      !r < !len && let c = !str.[!r] in
      (c>='0' && c<='9')
    in
      while blank_space () do incr r done;
      l := !r;
      lex := match !str.[!l] with
      | c when upper_lower () ->
          while alpha_num () do incr r done;
          K_Word  
      | c when ascii_num () ->
          num := 0;
          while ascii_num () do
            let d = int_of_char !str.[!r] - int_of_char '0'
            in  num := !num * 10 + d; incr r
          done;
          K_Number
      | c ->
          incr r;
          K_Mono

  and read_string () =
    while !r < !len && !str.[!r] <> !str.[!l] do
      incr r
    done;
    (!r < !len) && (incr r; true)
  and same_word s =
    String.length s = !r - !l &&
    let a = ref 0 and b = ref !l in
    while !b < !r && s.[!a] = !str.[!b] do incr a; incr b done;
    !b = !r
  and same_poly s =
    let ls  = !l + String.length s
    in (ls <= !len) &&
    let i = ref 0 and j = ref !l in
    while !j < ls && s.[!i] = !str.[!j] do
      incr i; incr j; 
    done;
    if !j = ls then r := ls;
    !j = ls

  and raise_denied demand =
    let position = {line = !row; text= !str; start = !l; stop = !r}
    in  raise (Demand_denied(position,demand))
  in

  (* demands *)
  let demand_line () =
    read_text (); read_next ()
  and demand_word s =
    match !lex with
    | K_Word when same_word s -> read_next ()
    | _ -> raise_denied (Word s)
  and demand_mono c =
    match !lex with
    | K_Mono when !str.[!l]=c -> read_next ()
    | _ -> raise_denied (Mono c)
  and demand_poly s =
    match !lex with
    | K_Mono when same_poly s -> read_next ()
    | _ -> raise_denied (Word s)
  and demand_name () =
    match !lex with
    | K_Word ->
      let s = String.sub !str !l (!r - !l) in read_next (); s
    | _ -> raise_denied Name
  and demand_not reserved =
    match !lex with
    | K_Word ->
      let s = String.sub !str !l (!r - !l) in
      if reserved s then begin
        let pos = {line = !row; text= !str; start = !l; stop = !r}
        in  raise (Word_reserved pos)
      end;
      read_next (); s
    | _ -> raise_denied Name
  and demand_string () =
    match !lex with
    | K_Mono when read_string () ->
      let s = String.sub !str !l (!r - !l) in read_next (); s
    | _ -> raise_denied String
  and demand_char () =
    match !lex with
    | K_Mono when (incr r; !r < !len)  && (!str.[!l] = !str.[!r]) ->
      incr r; !str.[succ !l] 
    | _ -> decr r; raise_denied Char
  and demand_int () =
    match !lex with
    | K_Number -> let n = !num in read_next (); n 
    | _ -> raise_denied Int
  (* promises *)
  and promised_word s =
    !lex = K_Word && same_word s
  and promised_mono c =
    !lex = K_Mono && !str.[!l]=c
  and promised_name () =
    !lex = K_Word 
  and promised_int () =
    !lex = K_Number 
  (* grants *)
  and granted_word s =
    match !lex with
    | K_Word when same_word s -> read_next (); true
    | _ -> false
  and granted_mono c =
    match !lex with
    | K_Mono when !str.[!l]=c -> read_next (); true
    | _ -> false
  and granted_poly s =
    match !lex with
    | K_Mono when same_poly s -> read_next (); true
    | _ -> false
  in

  demand_line ();
  {
  (* demands *)
  demand_line   = demand_line;
  demand_word   = demand_word;
  demand_mono   = demand_mono;
  demand_poly   = demand_poly;
  demand_name   = demand_name;
  demand_not    = demand_not;
  demand_string = demand_string;
  demand_char   = demand_char;
  demand_int    = demand_int;
  (* promises *)
  promised_word = promised_word;
  promised_mono = promised_mono;
  promised_name = promised_name;
  promised_int  = promised_int;
  (* grants *)
  granted_word  = granted_word;
  granted_mono  = granted_mono;
  granted_poly  = granted_poly;
  }


(* printers *)

let print_position p =
  if Sys.os_type="Unix" then begin
    print_string (String.sub p.text 0 p.start);
    print_string "\027[4m";
    print_string (String.sub p.text p.start (p.stop - p.start));
    print_string "\027[0m";
    print_string (String.sub p.text p.stop (String.length p.text - p.stop));
    print_string "\nline: ";
    print_int    p.line;
  end else begin
    print_string "line: ";
    print_int    p.line;
    print_string " | erratum: ";
    print_string (String.sub p.text p.start (p.stop - p.start));
  end

let print_denied demand =
  print_string " | expected: ";
  ( match demand with
  | Word str -> print_string "keyword "; print_string str
  | Mono chr -> print_string "character "; print_char chr
  | Name     -> print_string "an identifier"
  | String   -> print_string "a string "
  | Char     -> print_string "a character"
  | Int      -> print_string "an integer" );
  print_newline ()

let print_reserved () =
  print_string " | this identifier is a reserved word";
  print_newline ()

