
(* ocamlopt -unsafe -o TP5.exe Lex.ml TP5.ml *)

open Lex;;

let lex = Lex.make stdin in

let rec simple_name () =
  ignore(lex.demand_not reserved)
    
and reserved =
  Lex.sorted_array_mem
  [|
  "and";"array";"asm";
  "begin";"case";"const";
  "div";"do";"downto";"else";"end";
  "file";"for";"function";"goto";
  "if";"implementation";"in";"interface";
  "label";"mod";"nil";"not";"of";"or";"out";
  "packed";"procedure";"program";
  "record";"repeat";
  "set";"shl";"shr";"string";
  "then";"to";"type";
  "unit";"until";"uses";
  "var";"while";"with";"xor";			
  |]

and qualify_name () =
  ignore(lex.granted_mono '^');
  match lex with
  | l when lex.granted_mono '.' ->
        simple_name (); qualify_name ()
  | l when lex.granted_mono '[' ->
        expression ();
        lex.demand_mono ']';
        qualify_name ()
  | l -> arguments ()

and arguments () =
  if lex.granted_mono '(' then begin
    expression ();
    while lex.granted_mono ',' do 
      expression ()
    done;
    lex.demand_mono ')'; qualify_name ()
  end

and expression () =
  ( match lex with
  | l when lex.promised_int ()    ->
        ignore (lex.demand_int ());
        if lex.granted_mono '.' then ignore (lex.demand_int ())  
  | l when lex.promised_mono '\"' -> ignore (lex.demand_char ())
  | l when lex.promised_mono '\'' -> ignore (lex.demand_string ())
  | l when lex.granted_mono  '('  -> expression (); lex.demand_mono ')'
  | l when lex.granted_mono  '-'  -> expression ()
  | l when lex.granted_word "not" -> expression ()
  | l -> simple_name (); qualify_name ()
  ) ;
  ( match lex with
  | l when lex.granted_mono '+'   -> expression ()
  | l when lex.granted_mono '-'   -> expression ()
  | l when lex.granted_mono '*'   -> expression ()
  | l when lex.granted_poly "<="  -> expression ()
  | l when lex.granted_poly ">="  -> expression ()
  | l when lex.granted_poly "<>"  -> expression ()
  | l when lex.granted_mono '<'   -> expression ()
  | l when lex.granted_mono '>'   -> expression ()
  | l when lex.granted_mono '='   -> expression ()
  | l when lex.granted_word "and" -> expression ()  
  | l when lex.granted_word "or"  -> expression ()
  | l when lex.granted_word "xor" -> expression ()  
  | l when lex.granted_word "div" -> expression ()  
  | l when lex.granted_word "mod" -> expression ()
  | l -> ()
  ) 


and sequence () =
  lex.demand_word "begin";
  instruction ();
  while lex.granted_mono ';' do 
    instruction ()
  done;
  lex.demand_word "end"

and instruction () =
  match lex with
  | l when lex.promised_word "begin" ->
        sequence ();
  | l when lex.granted_word "if" ->
        expression ();
        lex.demand_word "then";
        instruction ();
        if lex.granted_word "else" then instruction ()
  | l when lex.granted_word "while" ->
        expression ();
        lex.demand_word "do";
        instruction ();        
  | l when lex.granted_word "repeat" ->
        instruction ();
        while lex.granted_mono ';' do 
          instruction ()
        done;
        lex.demand_word "until"; expression ()
  | l when lex.granted_word "for" ->
        simple_name (); qualify_name ();
        lex.demand_poly ":=";
        expression ();
        if lex.granted_word "downto" then ()
        else lex.demand_word "to";
        expression ();
        lex.demand_word "do";
        instruction ();        
  | l when lex.granted_word "with" ->
        simple_name (); qualify_name ();
        lex.demand_word "do";
        instruction ();        
  | l ->
        simple_name (); qualify_name ();
        if lex.granted_poly ":=" then expression ()
        else arguments()


and data_type () =
  match lex with
  | l when lex.granted_word "Integer" -> ()
  | l when lex.granted_word "Longint" -> ()
  | l when lex.granted_word "Real" -> ()
  | l when lex.granted_word "Byte" -> ()
  | l when lex.granted_word "Word" -> ()
  | l when lex.granted_word "Char" -> ()
  | l when lex.granted_word "Boolean" -> ()
  | l when lex.granted_word "Pointer" -> ()
  | l when lex.granted_word "String" ->
        if lex.granted_mono '[' then begin
          constant_value ();
          lex.demand_mono ']'
        end
  | l when lex.granted_word "file" ->
        lex.demand_word "of";
        data_type ()
  | l when lex.granted_word "array" ->
        lex.demand_mono '[';
        constant_value ();
        lex.demand_poly "..";
        constant_value ();
        lex.demand_mono ']';
        lex.demand_word "of";
        data_type ()
  | l when lex.granted_word "record" ->
        components ();
        lex.demand_word "end"
  | l when lex.granted_word "procedure" ->
        procedure_type ()
  | l when lex.granted_word "function" ->
        procedure_type (); lex.demand_mono ':'; data_type ()
  | l when lex.granted_mono '^' ->
        data_type ()
  | l when lex.granted_mono '(' ->
        simple_name ();
        while lex.granted_mono ',' do 
          simple_name ()
        done;
        lex.demand_mono ')'
  | l -> simple_name ()

and constant_value () =
  if lex.promised_name () then simple_name () 
  else ignore (lex.demand_int ())

and components () =
  simple_name ();
  while lex.granted_mono ',' do 
    simple_name ()
  done;
  lex.demand_mono ':';
  data_type ();
  if lex.granted_mono ';' then components ()

and procedure_type () =
  if lex.granted_mono '(' then
  let rec loop () =
    ( match lex with
    | l when lex.granted_word "const" -> ()
    | l when lex.granted_word "var" -> ()
    | l when lex.granted_word "in" -> ()
    | l when lex.granted_word "out" -> ()
    | l -> ()
    ) ;
    simple_name ();
    while lex.granted_mono ',' do 
      simple_name ()
    done;
    lex.demand_mono ':';
    data_type ();
    if lex.granted_mono ';' then loop ()
  in loop (); lex.demand_mono ')'


and procedure_body () =
  if not (lex.granted_word "forward") then begin 
    if lex.granted_word "var" then components ();
    sequence ();
  end;
  lex.demand_mono ';'

and procedure () =
  simple_name (); procedure_type ();
  lex.demand_mono ';';
  procedure_body ()

and typed_procedure () =
  simple_name (); procedure_type ();
  lex.demand_mono ':'; data_type (); lex.demand_mono ';';
  procedure_body ()

and constant () =
  comments ();
  simple_name ();
  lex.demand_mono '=';
  constant_value ();
  if lex.granted_mono ';' then constant ()

and definition () =
  comments ();
  simple_name ();
  lex.demand_mono '=';
  data_type ();
  if lex.granted_mono ';' then definition ()

and comments () =
  while lex.granted_poly "//" do
    lex.demand_line ()
  done

and declare () =
  comments ();
  match lex with
  | l when lex.granted_word "const" -> 
        comments (); constant (); declare ()
  | l when lex.granted_word "type" -> 
        comments (); definition (); declare ()
  | l when lex.granted_word "var" -> 
        comments (); components (); declare ()
  | l when lex.granted_word "procedure" -> 
        comments (); procedure (); declare ()
  | l when lex.granted_word "function" -> 
        comments (); typed_procedure (); declare ()
  | l -> ()

and program () =
  lex.demand_word "program";
  simple_name ();
  lex.demand_mono ';';
  comments ();
  declare ();
  sequence ()

and imports () =
  comments ();
  while lex.granted_word "uses" do
    comments ();
    simple_name ();
    while lex.granted_mono ',' do
      simple_name ()
    done;
    lex.demand_mono ';'
  done;

and modular () =
  lex.demand_word "unit";
  simple_name ();
  lex.demand_mono ';';
  comments ();
  lex.demand_word "interface";
  imports ();
  comments ();
  declare ();
  lex.demand_word "implementation";
  imports ();
  declare ();
  if lex.promised_word "begin" then sequence ()
  else lex.demand_word "end"

and turbo_pascal5 () =
  try
    comments ();
    if lex.promised_word "unit" then modular ()
    else program ();
    if lex.promised_mono '.' then ()
    else lex.demand_mono '.' 
  with
  | Lex.Demand_denied(position,demand) ->
      Lex.print_position position;
      Lex.print_denied   demand;
  | Lex.Word_reserved(position) ->
      Lex.print_position position;
      Lex.print_reserved ();
  | End_of_file ->
      print_endline "unexpected end of text"

in turbo_pascal5()
